#########################################
# This script is designed to load a 
# saved keras model and use it to make
# probabilistic predictions, compute the
# BSS and BS of those predictions, and 
# output an image of the forecast and 
# observations
#########################################

import numpy as np
import keras 
from keras.models import load_model 
import modelEvaluationFunctions_new as Eval
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import roc_curve, roc_auc_score, brier_score_loss
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import pickle 

#======================================
# Functions
#======================================

def unpickle(filename):
    f = open(filename, 'rb')
    new = pickle.load(f)
    f.close()

    return new


#Saves individual-day probabilities to .txt file 
#(saves over entire domain; not just mask points) 
#mask_probs is the probs within the analysis domain (1D) from a single day 
#ddx andd ddy are the number of x and y points, respectively in the full area (not just the mask) 
#mask_pts is an array of the points within the domain area (y,x)
#outdir is the output directory for the txt file
#fdate is the date in question
def save_probs(mask_probs, ddx, ddy, mask_pts, outdir, fdate):

    output_probs = np.ones((dy,dx))*-999.

    for l in range(len(mask_probs)):
        y_pt = mask_pts[l,0]
        x_pt = mask_pts[l,1]
        output_probs[y_pt, x_pt] = mask_probs[l]


    np.savetxt("%s/%s_dl_probs.txt" %(outdir, fdate), output_probs)

    return

def read_mask(infile, dx_pts, dy_pts, header):
    #@infile is the filename of unformatted binary file
    #Returns a numpy array of data
    #@dx_pts is an integer describing the number of longitude points
    #@dy_pts is an integer describing the number of latitude points
    #@header is a binary variable -- True if contains a 1-elmnt header/footer at 
    #beginning and end of file; else False


    f = open ( infile , 'rb' )
    arr = np.fromfile ( f , dtype = np.float32 , count = -1 )
    f.close()

    if (header == True):
        arr = arr[1:-1]


    #print arr.shape
    #Rebroadcast the data into 2-d array with proper format
    arr.shape = (dy_pts, dx_pts)
    return arr

#Returns an array containing (y,x) grid points in the mask AND
#an array containing (y,x) grid points outside the mask
#mask_var is a np array of binary mask points (1 in the mask; else 0)
#nx is total number of x grid points, ny is total num of y grid points
def get_domain_pts(mask_var, nx, ny):

    mask_pts = []
    out_pts = []
    for i in range(nx):
        for j in range(ny):
            if (mask_var[j,i] >= 0.9):
                mask_pts.append((j,i))
            else:
                out_pts.append((j,i))

    mask_pts = np.array(mask_pts)
    out_pts = np.array(out_pts)

    return mask_pts, out_pts


def read_binary(infile, dx_pts, dy_pts, dsamples, dchannels):
    #@infile is the filename of unformatted binary file
    #Returns a numpy array of data
    #@dx_pts is an integer describing the number of longitude points
    #@dy_pts is an integer describing the number of latitude points

    f = open ( infile , 'rb' )
    arr = np.fromfile ( f , dtype = np.float32 , count = -1 )
    f.close()

    #Rebroadcast the data into 2-d array with proper format
    arr.shape = (dsamples, dy_pts, dx_pts, dchannels)
    return arr


def read_obs_binary(infile):
    #@infile is the filename of unformatted binary file
    #Returns a numpy array of data

    f = open ( infile , 'rb' )
    arr = np.fromfile ( f , dtype = np.float32 , count = -1 )
    f.close()

    return arr

#Adds NaN values to dataset where appropriate so we can still use the height/width that we want
#Data outside the domain will just be represented as NaNs 
#@day_fcst is the day's forecast vector [shape(dy,dx, n_channels)]
#@dx and @dy are as given in the User defined params (scalar)
#@new_dx and @new_dy are the new values after padding (scalar)
#@n_chan is the number of channels (scalar)
def pad_dataset(day_fcst, dx, dy, new_dx, new_dy, n_chan):

    #initialize new array with all nans
    new_day_fcst = np.empty((new_dy, new_dx, n_chan))
    new_day_fcst[:] = np.nan

    #replace the real values from original forecast
    for c in range(n_chan):
        new_day_fcst[:dy, :dx, c] = day_fcst[:, :,c]


    return new_day_fcst



#Puts the 2d raw baseline data in proper 1-d format 
def get_raw_baseline_patch(probs_2d, mask_pts, size_patch):

    probs = np.zeros(len(mask_pts))

    for i in range(len(mask_pts)):
        ypt = mask_pts[i,0]
        xpt = mask_pts[i,1]
        probs[i] = probs_2d[ypt,xpt] 

    return probs 



#Returns a vector of the daily samples [shape (samples, height, width, channels)]
#and the daily observations [i.e., of shape (samples)]. 
#@day_fcst is the day's forecast vector [shape (dy,dx,n_channels)]
#@day_obs is the day's observation vector [shape (dy,dx)]
#@mask_pts is the set of domain points [shape (dy,dx)]
#@size_patch is the size of the width/height of the image (i.e., a scalar)
#@n_chan is the number of channels
#@samp_rate is the number of points to randomly sample
def get_patch_data(day_fcst, day_obs, mask_pts, size_patch, n_chan, samp_rate, llats, llons):

    if (samp_rate < 1.):
        n_rand_samples = int(len(mask_pts)*samp_rate)
    
    else:
        n_rand_samples = len(mask_pts)

    samples = np.zeros((n_rand_samples, size_patch, size_patch, n_chan))
    obs_list = np.zeros(n_rand_samples)
    point_radius = (size_patch)//2 #radius from center point


    #New approach: to limit the amount of samples (memory issues): maybe sample 10% of the points 
    #Step 1: Randomly sample the points in mask_pts 

    if (samp_rate < 1.):
        rand_inds = np.random.choice(np.arange(len(mask_pts)), size=n_rand_samples, replace=False)
        sample_mask_pts = mask_pts[rand_inds]

    else:
        sample_mask_pts = mask_pts

    #span the number of points in the analysis domain
    #for each point: get the obs and get the 101 x 101 x n-channels image
    for i in range(len(sample_mask_pts)):
        #set the x and y points
        #print (sample_mask_pts[i])

        ypt = sample_mask_pts[i,0]
        xpt = sample_mask_pts[i,1]
        #print (ypt, xpt)

        #get the obs
        obs_list[i] = day_obs[ypt, xpt]

        #get the lat and lon
        lat_field = llats[ypt-point_radius:ypt+point_radius+1, xpt-point_radius:xpt+point_radius+1]
        lon_field = llons[ypt-point_radius:ypt+point_radius+1, xpt-point_radius:xpt+point_radius+1]

        #get the forecast patch
        for c in range(n_chan):
            #samples[i, :,:, c] = day_fcst[ypt-point_radius:ypt+point_radius+1,xpt-point_radius:xpt+point_radius+1,c]
            #print (ypt, xpt)
            if (c < n_chan - 2): #if not last two channels:
                samples[i, :,:, c] = day_fcst[ypt-point_radius:ypt+point_radius+1,xpt-point_radius:xpt+point_radius+1,c]
            elif (c == n_chan -2): #the lat/lon fields will be the last two channels
                samples[i, :,:, c] = lat_field
            elif (c == n_chan -1 ):
                samples[i, :,:, c] = lon_field

    return samples, obs_list

#Saves a 2-d image with probabilities filled and binary observations contoured
#@nx is number of x points (i.e., dx)
#@ny is number of y points (i.e., dy)
#@mask_pts is an array (y,x) of points within the analysis domain
#@probs is the forecast probabilities - 1D
#@obs_mm is the observations in mm - 1D
#@thresh is the threshold in mm
def get_image(nx, ny, mask_pts, probs, obs_mm, thresh, outfile):
    #use_probs = probs.copy()   #to avoid overwriting probs 
    #use_obs = obs_mm.copy()
 
    #Initialize everything to -999 (equivalent to Nan). 
    obs_2d = np.ones((ny, nx))*-999.
    probs_2d = np.ones((ny,nx))*-999. 
  
    #Binarize observation
    #bin_obs = np.where(obs_mm >= thresh, 1., 0.) 

    bin_obs = obs_mm.copy() #Since already binarized

    #Fill the 2-d arrays with data in the mask
    for i in range(len(mask_pts)): 
        y = mask_pts[i,0]
        x = mask_pts[i,1]

        obs_2d[y,x] = bin_obs[i] #this should be correct, I think...
        probs_2d[y,x] = probs[i]

    #now can make the plot
    fig = plt.figure(figsize=(12,12), facecolor='whitesmoke')
    y = np.arange(ny)
    x = np.arange(nx)
    
    xx, yy = np.meshgrid(x,y) 

    #rng = np.arange(0.1, 1.1, 0.1) 
    #rng = np.array([0.02, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
    rng = np.array([0.02, 0.15, 0.30, 0.45, 0.60, 0.75, 0.90, 1.0]) 
    rng_obs = np.arange(0., 2., 1.)
    my_cmap = matplotlib.cm.rainbow
    filled_plt = plt.contourf(xx, yy, probs_2d, rng, cmap=my_cmap)
    contour_plt = plt.contour(xx, yy, obs_2d, rng_obs, colors='black', linewidths=2)
    cblabel = "Probability of Precip. > %s mm" %(str(thresh))
    cb = fig.colorbar(filled_plt, orientation='horizontal', label=cblabel)


    #restrict the plot domain 
    #ymin=43; ymax=190
    #xmin=76; xmax=328

    ymin=min(domain_pts[:,0]) 
    ymax=max(domain_pts[:,0])
    xmin=min(domain_pts[:,1])
    xmax=max(domain_pts[:,1])

    plt.xlim([xmin,xmax])
    plt.ylim([ymin,ymax])

    plt.savefig(outfile) 

   
#Saves a 2-d image of raw baseline probs given a 2-d field of raw baseline probabilities
#(that's the key difference between this method and the above method: probs is now 2d). 
#NOTE: Defunct now
def get_raw_baseline_image(nx, ny, mask_pts, probs_2d, obs_mm, thresh, outfile):
    
    #Get obs in proper format
    obs_2d = np.ones((ny, nx))*-999.
    
    #binarize observation
    bin_obs = np.where(obs_mm >= thresh, 1., 0.)
    
    for i in range(len(mask_pts)):
        y = mask_pts[i,0]
        x = mask_pts[i,1]

        obs_2d[y,x] = bin_obs[i]

    #now make the plot
    fig = plt.figure(figsize=(12,12), facecolor='whitesmoke')
    y = np.arange(ny)
    x = np.arange(nx)

    xx, yy = np.meshgrid(x,y)

    #rng = np.arange(0.1, 1.1, 0.1) 
    rng = np.array([0.02, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
    rng_obs = np.arange(0., 2., 1.)
    my_cmap = matplotlib.cm.rainbow
    filled_plt = plt.contourf(xx, yy, probs_2d, rng, cmap=my_cmap)
    contour_plt = plt.contour(xx, yy, obs_2d, rng_obs, colors='black', linewidths=2)
    cblabel = "Probability of Precip. > %s mm" %(str(thresh))
    cb = fig.colorbar(filled_plt, orientation='horizontal', label=cblabel)


    #restrict the plot domain 
    #ymin=43; ymax=190
    #xmin=76; xmax=328

    ymin=min(domain_pts[:,0])
    ymax=max(domain_pts[:,0])
    xmin=min(domain_pts[:,1])
    xmax=max(domain_pts[:,1])


    plt.xlim([xmin,xmax])
    plt.ylim([ymin,ymax])


    plt.savefig(outfile)



 

#Will print an image that displays 100% probs where the ensemble mean forecast
#meets or exceeds the threshold and 0% elsewhere. Obs contours will still be in
#black
#Note: probs and obs are already in 2D. Just need to binarize the obs and restrict
#the domain
def get_simple_baseline_image(nx, ny, mask_pts, probs, obs_mm, thresh, outfile):


    #bin_obs = np.where(obs_mm >= thresh, 1., 0.) 
    bin_obs = obs_mm.copy()
    bin_probs = np.where(probs >= thresh, 1., 0.)
    
    fig = plt.figure(figsize=(12,12), facecolor='whitesmoke')
    y = np.arange(ny)
    x = np.arange(nx)

    xx, yy = np.meshgrid(x,y)

    rng = np.arange(0.1, 1.1, 0.1) 
    rng_obs = np.arange(0., 2., 1.)
    my_cmap = matplotlib.cm.rainbow
    filled_plt = plt.contourf(xx, yy, bin_probs, rng, cmap=my_cmap)
    contour_plt = plt.contour(xx, yy, bin_obs, rng_obs, colors='black', linewidths=2)
    cblabel = "Probability of Precip. > %s mm" %(str(thresh))
    cb = fig.colorbar(filled_plt, orientation='horizontal', label=cblabel)


    #restrict the plot domain 
    #ymin=43; ymax=190
    #xmin=76; xmax=328

    ymin=min(domain_pts[:,0])
    ymax=max(domain_pts[:,0])
    xmin=min(domain_pts[:,1])
    xmax=max(domain_pts[:,1])


    plt.xlim([xmin,xmax])
    plt.ylim([ymin,ymax])


    plt.savefig(outfile)

#Formats dataset for the random forest
#i.e., converts from (samples, y, x, channels)
#to (samples, variables) 
def format_for_rf(in_data, in_mean, in_std, nsize):

    #First, need to "un-preprocess" in_data
    use_data = in_data.copy()
    use_data *= in_std
    use_data += in_mean 


    patch_size = len(use_data[0,:,0,0])
    n_samples = len(use_data[:,0,0,0])
    n_channels = len(use_data[0,0,0,:])

    radius = nsize // 2 #radius from center point
    center_pt = patch_size // 2

    #variables will be as follows:
    #channel1 at -5 to +5 from center point, channel 2 at -5 to +5, etc. 

    n_vars = nsize*nsize*n_channels #total # of vars considered by sklearn
    #new_data = np.zeros((n_samples, n_vars))

    #Extract just the relevant "patches"
    new_data = use_data[:,center_pt-radius:center_pt+radius+1, center_pt-radius:center_pt+radius+1, :]

    #reshape to a 2d array

    #order: spans through the variables first and then the points
    new_data = new_data.reshape(n_samples, -1)

    return new_data


#==========================================================
# Get aggregate stats using only the testing data forecasts
#==========================================================
def getAggStats(probabilities, Y_eval):
    #numForecasts is the number of forecasts in each probability bin
    RELx, RELy, numForecasts = Eval.ReliabilityCurve( probabilities, Y_eval)
    POFD, POD = Eval.rocCurve(probabilities, Y_eval )

    JUNK, SR = Eval.perfDiag(probabilities, Y_eval)

    AUC = roc_auc_score(Y_eval, probabilities)

    BSS, BS = Eval.bss_no_train(Y_eval, probabilities)

    BSS_rel, BSS_res, BSS_unc = Eval.BS_decomposition(Y_eval, probabilities)

    return POD, POFD, SR, AUC, RELx, RELy, BSS, BSS_rel, BSS_res, BSS_unc, BS, numForecasts


#======================================
# User defined parameters
#======================================

threshold = 2.54 #NOTE: Not used (or shouldn't be) 

#dx = 369
#dy = 257

dx = 93
dy = 65

n_channels = 12
patch_size = 5

n_size_rf = 5 #consider points from 5x5 area

saved_model_file = 'separable_conv_tutorial_5x5.h5'

samp_rate = 1.0
n_mask_pts = 1434
samp_per_day = int(n_mask_pts*samp_rate)

#fdate_file = "ncar_sseo_dates.txt"
#odate_file = "ncar_sseo_obs_dates.txt"

#mask_file = "great_plains_domain_pts.npy" 
#mask_file = "conus_mask_pts_80km.npy"

mask_file = "new_conus_mask_80km.dat"
#fcst_dir = "/gpfs/fs1/scratch/eloken/ncar/ncar_ensemble/20km_files/temporal_mean"

#fcst_dir = "/glade/scratch/eloken/sseo/sseo_files/80km_files/ens_stats"

fcst_dir = "./fcst"

#obs_date_file = "ncar_sseo_grib_obs_dates.txt"
#obs_dir = "/gpfs/fs1/work/eloken/stage4_verification/20km_dat_files/24h"
#obs_dir = "/glade/work/eloken/storm_reports/all_hazards"

obs_dir = "./obs"

#directory to raw baseline probabilities
#raw_dir = "/gpfs/fs1/scratch/eloken/ncar/ncar_ensemble/grib_files/full_period/temporal_sum/32bit_data/20km/raw_baseline"
#raw_baseline_dir = "/glade/scratch/eloken/sseo/sseo_files/20km_files/raw_baseline"

variables = ["REFDHM", "RELH", "TMPK", "DWPK", "P01M", "UPDRHM", "CAPE", "SHEAR", "UPHLHM"]

#fhrs = ["f012", "f024", "f036"]

mean_file = "train_mean.npy"
std_file = "train_std.npy"

test_date_file = "test_dates.txt"
train_date_file = "train_dates.txt"
val_date_file = "val_dates.txt"

outdir_txt = "./outfiles"

sampling_rate = 1.1 #greater than 1 b/c we want the whole patch

#set appropriate missing fill values -- will depend on channel
#order of variables: DPTf12-36, MSLPf12-36, TMPf12-36, APCP, CAPE, PWAT
#200 for DPT and TMP; 900000 for MSLP; -1 for APCP, PWAT, and CAPE
#missing_fill_values = [200., 200., 200., 90000., 90000., 90000., 200., 200., 200., -1., -1., -1.]

#======================================

#Read in the dates
test_dates = np.genfromtxt(test_date_file, dtype='str') 

#Split dates (and data) into testing, training, validation sets

#Read in lats/lons
lats = read_mask("lat_80km.dat", dx, dy, True)
lons = read_mask("lon_80km.dat", dx, dy, True)


#Read in the mask points 
#domain_pts = np.load(mask_file) #(ypts, xpts) 

domain_mask = read_mask(mask_file, dx, dy, True)

#Obtain x and y points/coordinates that fall within the domain
domain_pts, non_domain_pts = get_domain_pts(domain_mask, dx, dy)

samples_per_day = len(domain_pts)


#Load preprocessing data
preproc_mean = np.load(mean_file)
preproc_std = np.load(std_file)

#NOTE: In the future, won't have to do this
#delta = 0.000001
#preproc_std += delta

#Load saved keras model
#keras_model = load_model("simple_model_2.54mm.h5")
#keras_model = load_model("residual_model_2.54mm.h5")
#keras_model = load_model("xception_model.h5")
#keras_model = load_model("unet_model.h5")

keras_model = load_model(saved_model_file)

#Get the testing data in a format that can be read in by the keras model (which will be loaded later)
for d in range(len(test_dates)):
    daily_fcst = np.zeros((dy,dx, n_channels))
    fdate = test_dates[d]
    #odate = test_odates[d]
    #odate = fdate #NOTE: this is true here b/c of SPC naming convention
    print (fdate)

    
    #ffiles = ["%s/sseo_ens_%s_%s_80km_%s00.dat" %(fcst_dir, var, fdate) for var in variables]

    #ofile = "%s/%s00_REPS211.dat" %(obs_dir, odate)

    #Read in the forecast and obs data and the raw baseline data
    #daily_obs = read_binary(ofile, dx, dy, False)
    #for c in range(len(ffiles)): 
    #    daily_fcst[:,:,c] = read_binary(ffiles[c], dx, dy, False)


    testing_x = read_binary('%s/%s_fpatch.dat' %(fcst_dir, fdate), patch_size, patch_size, samp_per_day, n_channels)

    testing_y = read_obs_binary('%s/%s_obs.dat' %(obs_dir, fdate))

    #Implement the patches algorithm to obtain each day's samples
    #daily_samples, daily_patch_obs = get_patch_data(daily_fcst, daily_obs, domain_pts, patch_size, n_channels, sampling_rate, lats, lons)

    #print ("after get_patch")
   
    #Next, need to do preprocessing -- same as before: subtract the mean (from training set) and divide by standard deviation
    testing_x -= preproc_mean
    testing_x /= preproc_std 

    #TODO: Would impute missing values to 0s here (i.e., missing vals become mean vals)

    #Now, use model to make predictions

    daily_predictions = keras_model.predict(testing_x)[:,0]

    #TODO: Obtain probability text file

    #daily_predictions = 1-daily_predictions

    #TODO: print out an image

    out_filename = "%s_probs.txt" %(fdate)


    save_probs(daily_predictions, dx, dy, domain_pts, outdir_txt, fdate)



