##############################################################
# This file does preprocessing for Deep Learning
##############################################################


import numpy as np
import os
from sklearn.metrics import roc_curve, roc_auc_score, brier_score_loss
from sklearn.calibration import calibration_curve
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
import pickle


#==============================
# Functions
#==============================


#=============================
# Get the current date chunk
#=============================

def getDateChunk(in_dates,k, curr_chunk,ndays1,ndays2,nChunks1,nChunks2, daySplit):
# in_dates is the incoming array of dates (an array of strings)
# k is the total number of chunks
# curr_chunk is the current chunk for testing
# ndays1 is the number of days in chunk/group 1
# ndays2 is the number of days in chunk/group 2
# nChunks1 is the number of chunks in group1
# nChunks2 is the number of chunks in group2
# daySplit is the list of day indices telling how in_dates should be split

#Split the dates array into chunks. The first 4 chunks have 29 days;
#The last 3 chunks have 30 days

#Obtain testing, validation, and training dates for current fold.
#How this works: testing is the current "chunk", 
#validation is the next "chunk", and training comprises all other chunks

    date_chunks = np.array_split(in_dates, daySplit)

    test_date = date_chunks[curr_chunk]

    #if the last chunk is the testing chunk
    if (curr_chunk == k-1):
        #val_date = date_chunks[0]
        train_date = np.concatenate(date_chunks[0:curr_chunk],axis=0)

    #if first chunk is the testing chunk
    elif (curr_chunk == 0):
        train_date = np.concatenate(date_chunks[1:],axis=0)

    else:
        #val_date = date_chunks[curr_chunk + 1]
        train_date1 = date_chunks[:curr_chunk]
        train_date2 = date_chunks[curr_chunk + 1:]
        train_date = train_date1 + train_date2
        train_date = np.concatenate((train_date), axis=0)


    return test_date, train_date


#====================================
# Get necessary info to make splits
#====================================

#Will obtain the list which shows how the data should be split
#(given uneven chunks)
def get_split_info(sampPerChunk1, sampPerChunk2,n1,n2):
#sampPerChunk1 is the samples_per_chunk1 variable (user defined)
#sampPerChunk2 is the samples_per_chunk2 variable
#n1 is the number of chunks in group 1
#n2 is the number of chunks in group 2
    chunk_list = []

    #first chunk
    for c in range(1,n1+1):
        chunk_list.append(c*sampPerChunk1)

    if (n2 > 0): #make sure we indeed have two chunks; if not, the follwoing code doesn't make sense
    #second chunk
        for c in range(1,n2): # only n2 instead of n2+1 b/c of how np.array_split works
            chunk_list.append(n1*sampPerChunk1+c*sampPerChunk2)

    return chunk_list



def read_binary(infile, dx_pts, dy_pts, header):
    #@infile is the filename of unformatted binary file
    #Returns a numpy array of data
    #@dx_pts is an integer describing the number of longitude points
    #@dy_pts is an integer describing the number of latitude points
    #@header is a binary variable -- True if contains a 1-elmnt header/footer at 
    #beginning and end of file; else False


    f = open ( infile , 'rb' )
    arr = np.fromfile ( f , dtype = np.float32 , count = -1 )
    f.close()

    if (header == True):
        arr = arr[1:-1]


    #print arr.shape
    #Rebroadcast the data into 2-d array with proper format
    arr.shape = (dy_pts, dx_pts)
    return arr

def read_obs_binary(infile):
    #@infile is the filename of unformatted binary file
    #Returns a numpy array of data

    f = open ( infile , 'rb' )
    arr = np.fromfile ( f , dtype = np.float32 , count = -1 )
    f.close()

    return arr

#Returns an array containing (y,x) grid points in the mask 
#mask_var is a np array of binary mask points (1 in the mask; else 0)
#nx is total number of x grid points, ny is total num of y grid points
def get_domain_pts(mask_var, nx, ny):

    mask_pts = []
    for i in range(nx):
        for j in range(ny):
            if (mask_var[j,i] >= 0.9):
                mask_pts.append((j,i))

    mask_pts = np.array(mask_pts)

    return mask_pts

#Reformats the data to a format acceptable to sklearn
#i.e., going from (samples, y, x, channels) to (samples, variables)
#nsize is neighborhood size (e.g., 5 means consider 5x5 neighborhood around the center point)
def reformat_data(in_data, nsize, n_samples, fdate, center_pt, radius):

    #variables will be as follows:
    #channel1 at -5 to +5 from center point, channel 2 at -5 to +5, etc. 

    #n_vars = nsize*nsize*n_channels + 12#total # of vars considered by sklearn
    #new_data = np.zeros((n_samples, n_vars))

    #Extract just the relevant "patches"
    #new_data = in_data[:,center_pt-radius:center_pt+radius+1, center_pt-radius:center_pt+radius+1, 0:-2]
    
    new_data = in_data[:,center_pt-radius:center_pt+radius+1, center_pt-radius:center_pt+radius+1, :]

    #But how to handle the lat/lon -- only really want lat/lon at the center point
    #lats = in_data[:,center_pt, center_pt, -2]
    #lons = in_data[:, center_pt, center_pt, -1] 

    #lats.shape = (len(lats), 1) 
    #lons.shape = (len(lons), 1) 

    #reshape to a 2d array

    #order: spans through the variables first and then the points
    new_data = new_data.reshape(n_samples, -1)


    #Add the lats and lons
    #new_data = np.concatenate( (new_data, lats), axis=1)
    #new_data = np.concatenate((new_data, lons), axis=1)

    
    return new_data


#Returns a vector of the daily samples [shape (samples, height, width, channels)]
#and the daily observations [i.e., of shape (samples)]. 
#@day_fcst is the day's forecast vector [shape (dy,dx,n_channels)]
#@day_obs is the day's observation vector [shape (dy,dx)]
#@mask_pts is the set of domain points [shape (dy,dx)]
#@size_patch is the size of the width/height of the image (i.e., a scalar)
#@n_chan is the number of channels
#@samp_rate is the number of points to randomly sample
#@llats is the latitudes (j,i)
#@llons is the longitudes (j,i)

#TODO: implement samp_rate
def get_patch_data(day_fcst, day_obs, mask_pts, size_patch, n_chan, samp_rate, llats, llons):

    if (samp_rate < 1.):
        n_rand_samples = int(len(mask_pts)*samp_rate)
    else:
        n_rand_samples = len(mask_pts)

    samples = np.zeros((n_rand_samples, size_patch, size_patch, n_chan))
    obs_list = np.zeros(n_rand_samples)
    point_radius = (size_patch)//2 #radius from center point

    #New approach: to limit the amount of samples (memory issues): maybe sample 10% of the points 
    #Step 1: Randomly sample the points in mask_pts 

    if (samp_rate < 1.):
        rand_inds = np.random.choice(np.arange(len(mask_pts)), size=n_rand_samples, replace=False)
        sample_mask_pts = mask_pts[rand_inds]

    else:
        sample_mask_pts = mask_pts

    #span the number of points in the analysis domain
    #for each point: get the obs and get the 101 x 101 x n-channels image
    for i in range(len(sample_mask_pts)):
        #set the x and y points
        #print (sample_mask_pts[i])

        ypt = sample_mask_pts[i,0]
        xpt = sample_mask_pts[i,1]
        #print (ypt, xpt)

        #get the obs
        obs_list[i] = day_obs[ypt, xpt]

        #get the lat and lon
        lat_field = llats[ypt-point_radius:ypt+point_radius+1, xpt-point_radius:xpt+point_radius+1]
        lon_field = llons[ypt-point_radius:ypt+point_radius+1, xpt-point_radius:xpt+point_radius+1]


        #Original (i.e., DON'T express surrounding grid points in relative framework:
        #get the forecast patch
        for c in range(n_chan):
            #print (ypt, xpt)
            if (c < n_chan - 2): #if not last two channels:
                samples[i, :,:, c] = day_fcst[ypt-point_radius:ypt+point_radius+1,xpt-point_radius:xpt+point_radius+1,c]
            elif (c == n_chan -2): #the lat/lon fields will be the last two channels
                samples[i, :,:, c] = lat_field
            elif (c == n_chan -1 ):
                samples[i, :,:, c] = lon_field

        #New (i.e., DO express surrounding grid points in relative framework: -- NOTE: Didn't work as well
        #for c in range(n_chan):
        #    if (c < n_chan - 2): #if not last two channels:
        #        samples[i, :,:, c] = day_fcst[ypt-point_radius:ypt+point_radius+1,xpt-point_radius:xpt+point_radius+1,c] - day_fcst[ypt, xpt, c]
        #        #set the center point (b/c now will be zero)
        #        samples[i,point_radius, point_radius,c] = day_fcst[point_radius, point_radius, c]
        #    elif (c == n_chan -2): #the lat/lon fields will be the last two channels
        #        samples[i, :,:, c] = lat_field
        #    elif (c == n_chan -1 ):
        #        samples[i, :,:, c] = lon_field



    return samples, obs_list


#Does the preprocessing: outputs individual-day .dat files to 
#sepcified directory to be used later for kfold cross-validation
#centerPt is the center point of the patch 
#rad is the RF radius 
def do_preprocessing(fdates, ddx, ddy, n_chan, var_list, fdir, odir, domain_pts, size_patch, samp_rate, lats, lons, n_samples, nsize, outdir_fcst, outdir_obs, centerPt, rad):

    for d in range(len(fdates)):

        #daily_fcst = np.zeros((n_mask_pts, n_vars))
        daily_fcst = np.zeros((ddy, ddx, n_chan)) 
        fdate = fdates[d]
        print (fdate) 

        ffiles = ["%s/sseo_ens_mean_%s_80km_%s00.dat" %(fdir, var, fdate) for var in var_list] 

        ofile = "%s/%s00_REPS211_hail.dat" %(odir, fdate)

        #Read in the forecast and obs data
        daily_obs = read_binary(ofile, ddx, ddy, False)

        for c in range(len(ffiles)):
            daily_fcst[:,:,c] = read_binary(ffiles[c], ddx, ddy, False)

        print ("after read daily forecast")

        #Implement the patches algorithm to obtain each day's samples
        daily_samples, daily_patch_obs = get_patch_data(daily_fcst, daily_obs, domain_pts, size_patch, n_chan, samp_rate, lats, lons)

        print ("after get_patch")
        print (daily_samples.shape)
        print (daily_patch_obs.shape)

        #TODO: Now is when you'd want to reformat the data; then you can save it to .dat file
        #NOTE: We skip this step for deep learning 
        #new_daily_samples = reformat_data(daily_samples, nsize, n_samples, fdate, centerPt, rad)

        #daily_patch_obs is already in the proper format (shape = samples) 

        print (daily_samples.shape) 
        print (daily_patch_obs.shape) 
        new_daily_samples = np.float32(daily_samples)
        daily_patch_obs = np.float32(daily_patch_obs)

        #Write to file 
        new_daily_samples.tofile("%s/%s_fpatch.dat" %(outdir_fcst, fdate))
        daily_patch_obs.tofile("%s/%s_obs.dat" %(outdir_obs, fdate) )

    return

#Saves individual-day probabilities to .txt file 
#(saves over entire domain; not just mask points) 
#mask_probs is the probs within the analysis domain (1D) from a single day 
#ddx andd ddy are the number of x and y points, respectively in the full area (not just the mask) 
#mask_pts is an array of the points within the domain area (y,x)
#outdir is the output directory for the txt file
#fdate is the date in question
def save_probs(mask_probs, ddx, ddy, mask_pts, outdir, fdate):

    output_probs = np.ones((dy,dx))*-999.
    
    for l in range(len(mask_probs)):
        y_pt = mask_pts[l,0]
        x_pt = mask_pts[l,1]
        output_probs[y_pt, x_pt] = mask_probs[l]


    np.savetxt("%s/%s_rf_probs.txt" %(outdir, fdate), output_probs)

    return 



#==================================
# User defined parameters
#==================================

dx = 93 #num x points
dy = 65 #num y points

variables = ["REFDHM", "RELH", "TMPK", "DWPK", "P01M", "UPDRHM", "CAPE", "SHEAR", "UPHLHM", "CAPESHEAR" ]

date_file = "all_dates.txt"

n_mask_pts = 1434
n_size_rf = 5 #Look at 5x5 patches
n_vars = (len(variables)+2)*n_size_rf*n_size_rf # +2 because adding latitude and longitude

n_channels = len(variables)+2

sampling_rate = 1.0

radius = n_size_rf//2 #radius from center point
patch_size = n_size_rf
center_pt = patch_size//2

n_days = 221

samples_per_day = int(n_mask_pts*sampling_rate)
total_samples = int(n_days*samples_per_day)

maskfile = "new_conus_mask_80km.dat"

lat_file = "lat_80km.dat"
lon_file = "lon_80km.dat"

indir_forecast = "../all_x"
indir_observations = "../all_y"

outdir_forecast = "./fcst"
outdir_observations = "./obs"

#Map Projections

#NCEP 211 stats
#deltax = 81.2705*10**3 # horizontal grid spacing in m; 81.2705 km
#sw_lon = -133.459
#sw_lat = 12.190
#ne_lon = -49.385
#ne_lat = 57.290
#tlat1 = 25.0 #lat1
#tlat2 = 25.0 #lat2 
#stlon = -95.0  #lon_0

#====================================================

#Step 1: Read in the dates, lats, lons; get mask points, etc. 

dates = np.genfromtxt(date_file, dtype='str')

#Read in the 20km lat/lon points
lats = read_binary(lat_file, dx, dy, True)
lons = read_binary(lon_file, dx, dy, True)

domain_mask = read_binary(maskfile, dx, dy, True)

#Obtain x and y points/coordinates that fall within the domain
domain_pts = get_domain_pts(domain_mask, dx, dy)

#Step 2: Do the preprocessing -- get individual data .dat files for ALL DATES
do_preprocessing(dates, dx, dy, n_channels, variables, indir_forecast, indir_observations, domain_pts, patch_size, sampling_rate, lats, lons, samples_per_day, n_size_rf, outdir_forecast, outdir_observations, center_pt, radius)










